from io import StringIO, BytesIO
import os
try:
    from urlparse import urljoin
except ImportError:
    from urllib.parse import urljoin


from django.conf import settings
from django.core.files.storage import Storage
from django.core.files.base import File
try:
    from django.utils.encoding import filepath_to_uri, force_bytes
except ImportError:
    # django.utils.encoding.force_bytes doesn't exist in < 1.5
    from django.utils.encoding import smart_str as force_bytes
import six

from django.core.files.storage import Storage
from django.core.files.base import ContentFile
from django.utils.deconstruct import deconstructible
from django.utils.encoding import filepath_to_uri
from django.utils.six.moves.urllib.parse import urljoin
from django.utils import timezone


class PathDoesNotExist(Exception):
    pass


class InMemoryNode(object):
    """
    Base class for files and directories.
    """
    parent = None

    def add_child(self, name, child):
        child.parent = self
        self.children[name] = child


class InMemoryFile(InMemoryNode, File):
    """
    Stores contents of file and stores reference to parent. File interface is identical
    to ContentFile, except that self.size works even after data has been written to it
    """
    def __init__(self, content='', parent=None, name=None):
        #init InMemoryNode
        self.parent = parent
        self.created_at = timezone.now()
        self.last_modified = timezone.now()
        self.last_accessed = timezone.now()

        #init File
        if six.PY3:
            stream_class = StringIO if isinstance(content, six.text_type) else BytesIO
        else:
            stream_class = six.BytesIO
            content = force_bytes(content)
        File.__init__(self, stream_class(content), name=name)

    def __str__(self):
        return '<InMemoryFile: %s>' % self.name

    def __bool__(self):
        return True

    def __nonzero__(self):      # Python 2 compatibility
        return type(self).__bool__(self)

    @property
    def _size(self):
        pos = self.file.tell()
        self.file.seek(0, os.SEEK_END)
        size = self.file.tell()
        self.file.seek(pos)
        return size

    def open(self, mode=None):
        self.seek(0)

    def close(self):
        pass


class InMemoryDir(InMemoryNode):
    """
    Stores dictionary of child directories/files and reference to parent.
    """
    def __init__(self, dirs=None, files=None, parent=None):
        self.children = {}
        self.parent = parent

    def resolve(self, path, create=False, use_bytes=False):
        path = os.path.normpath(path)
        path_bits = path.strip(os.sep).split(os.sep, 1)
        current = path_bits[0]
        rest = path_bits[1] if len(path_bits) > 1 else None
        if not rest:
            if current == '.' or current == '':
                return self
            if current in self.children.keys():
                return self.children[current]
            if not create:
                raise PathDoesNotExist()
            content = six.binary_type() if use_bytes else six.text_type()
            node = InMemoryFile(name=current, content=content)
            self.add_child(current, node)
            return node
        if current in self.children.keys():
            return self.children[current].resolve(rest, create=create, use_bytes=use_bytes)
        if not create:
            raise PathDoesNotExist()
        node = InMemoryDir()
        self.add_child(current, node)
        return self.children[current].resolve(rest, create=create, use_bytes=use_bytes)

    def ls(self, path=''):
        return list(self.resolve(path).children.keys())

    def listdir(self, dir):
        nodes = tuple(six.iteritems(self.resolve(dir).children))
        dirs = [k for (k, v) in nodes if isinstance(v, InMemoryDir)]
        files = [k for (k, v) in nodes if isinstance(v, InMemoryFile)]
        return [dirs, files]

    def delete(self, path):
        node = self.resolve(path)
        for name, child in six.iteritems(node.parent.children):
            if child is node:
                del node.parent.children[name]
                break

    def exists(self, name):
        try:
            self.resolve(name)
        except PathDoesNotExist:
            return False
        else:
            return True

    def size(self, name):
        return self.resolve(name).size

    def open(self, path, mode="r"):
        create = "w" in mode
        use_bytes = "b" in mode
        f = self.resolve(path, create=create, use_bytes=use_bytes)
        f.open(mode)
        f.last_accessed = timezone.now()
        return f

    def save(self, path, content):
        mode = 'wb' if isinstance(content, six.binary_type) else 'w'
        with self.open(path, mode) as f:
            f.write(content)
        f.last_modified = timezone.now()
        return path


_filesystem = InMemoryDir()


@deconstructible
class InMemoryStorage(Storage):
    """
    Django storage class for in-memory filesystem.
    """
    def __init__(self, filesystem=None, base_url=None):
        if not filesystem and getattr(settings, "INMEMORYSTORAGE_PERSIST", False):
            self.filesystem = _filesystem
        else:
            self.filesystem = filesystem or InMemoryDir()

        if base_url is None:
            base_url = settings.MEDIA_URL
        self.base_url = base_url

    def deconstruct(self):
        pass

    def listdir(self, dir):
        return self.filesystem.listdir(dir)

    def delete(self, path):
        return self.filesystem.delete(path)

    def exists(self, name):
        return self.filesystem.exists(name)

    def size(self, name):
        return self.filesystem.size(name)

    def _open(self, name, mode="r"):
        return self.filesystem.open(name, mode)

    def _save(self, name, content):
        return self.filesystem.save(name, content.read())

    def url(self, name):
        if self.base_url is None:
            raise ValueError("This file is not accessible via a URL.")
        return urljoin(self.base_url, filepath_to_uri(name))

    def modified_time(self, name):
        file = self.filesystem.resolve(name)
        return file.last_modified

    def accessed_time(self, name):
        file = self.filesystem.resolve(name)
        return file.last_accessed

    def created_time(self, name):
        file = self.filesystem.resolve(name)
        return file.created_at



import time
import unittest
from inmemorystorage import InMemoryStorage
from inmemorystorage.storage import InMemoryDir, InMemoryFile
from django.core.files.base import ContentFile
from django.conf import settings
from django.test.utils import override_settings


class MemoryStorageTests(unittest.TestCase):
    def setUp(self):
        if not settings.configured:
            settings.configure(MEDIA_URL='')
        self.storage = InMemoryStorage()
        self.filesystem = self.storage.filesystem

    def test_listdir(self):
        self.assertEqual(self.storage.listdir(''), [[], []])

        self.filesystem.add_child('dir0', InMemoryDir())
        self.filesystem.add_child('file0', InMemoryFile())

        self.assertEqual(self.storage.listdir(''), [['dir0'], ['file0']])
        self.assertEqual(self.storage.listdir('dir0'), [[], []])

        self.filesystem.resolve('dir0').add_child('subdir', InMemoryDir())
        self.assertEqual(self.storage.listdir('dir0'), [['subdir'], []])
        self.assertEqual(self.storage.listdir('dir0/subdir'), [[], []])

    def test_delete(self):
        self.filesystem.add_child('dir0', InMemoryDir())
        self.filesystem.resolve('dir0').add_child('nested_file', InMemoryFile())
        self.filesystem.add_child('file0', InMemoryFile())
        self.assertEqual(self.filesystem.resolve('dir0').ls(), ['nested_file'])

        self.storage.delete('dir0/nested_file')
        self.assertEqual(self.filesystem.resolve('dir0').ls(), [])
        self.assertEqual(set(self.filesystem.ls()), set(['dir0', 'file0']))

        self.storage.delete('dir0')
        self.assertEqual(set(self.filesystem.ls()), set(['file0']))

    def test_exists(self):
        self.filesystem.add_child('file0', InMemoryFile())
        self.assertTrue(self.storage.exists('file0'))
        self.assertFalse(self.storage.exists('file1'))
        self.storage.delete('file0')
        self.assertFalse(self.storage.exists('file0'))

    def test_size(self):
        self.filesystem.add_child('file0', InMemoryFile('test'))
        self.assertEqual(self.storage.size('file0'), 4)

    def test_save(self):
        self.storage.save('file', ContentFile('test'))
        self.assertEqual(self.storage.size('file'), 4)
        self.storage.save('subdir/file', ContentFile('test'))
        self.assertEqual(self.storage.size('subdir/file'), 4)
        self.storage.save('bytes_file', ContentFile(b'test'))
        self.assertEqual(self.storage.size('bytes_file'), 4)

    def test_write(self):
        with self.storage.open("file", "w") as f:
            f.write("hello")
        with self.storage.open("file", "r") as f:
            self.assertEqual(f.read(), "hello")

    def test_all(self):
        self.assertEqual(self.storage.listdir('/'), [[], []])
        self.assertEqual(self.storage.save('dir/subdir/file', ContentFile('testing')), 'dir/subdir/file')
        self.assertEqual(self.storage.listdir('/'), [['dir'], []])
        self.assertEqual(self.storage.listdir('dir/'), [['subdir'], []])
        self.assertEqual(self.storage.save('dir/subdir/file2', ContentFile('testing2')), 'dir/subdir/file2')
        self.assertEqual(self.storage.save('file', ContentFile('testing3')), 'file')
        self.assertEqual(self.storage.listdir('/'), [['dir'], ['file']])
        self.assertEqual(self.storage.listdir('dir/'), [['subdir'], []])
        self.assertEqual(self.storage.open('dir/subdir/file').read(), 'testing')
        self.assertEqual(self.storage.size('dir/subdir/file'), 7)
        self.assertEqual(self.storage.size('dir/subdir/file2'), 8)
        self.assertEqual(self.storage.size('file'), 8)
        self.assertEqual(self.storage.delete('file'), None)
        self.assertEqual(self.storage.listdir('/'), [['dir'], []])
        self.assertEqual(self.storage.delete('dir/subdir/file'), None)
        self.assertEqual(self.storage.listdir('dir/subdir'), [[], ['file2']])
        self.assertEqual(self.storage.exists('dir/subdir/file2'), True)
        self.assertEqual(self.storage.delete('dir/subdir/file2'), None)
        self.assertEqual(self.storage.exists('dir/subdir/file2'), False)
        self.assertEqual(self.storage.listdir('dir/subdir'), [[], []])

    @override_settings(MEDIA_URL=None)
    def test_url(self):
        # main storage not instantiated with a base url so should raise not
        # implemented.
        storage = InMemoryStorage()

        self.assertRaises(ValueError, storage.url, ('file0',))

        storage = InMemoryStorage(base_url='http://www.example.com')

        self.assertEqual(storage.url('file0'), 'http://www.example.com/file0')

    def test_modified_time(self):
        self.storage.save('file', ContentFile('test'))
        modified_time = self.storage.modified_time('file')

        self.storage.delete('file')

        time.sleep(0.1)

        self.storage.save('file', ContentFile('test-again'))
        new_modified_time = self.storage.modified_time('file')
        self.assertTrue(new_modified_time > modified_time)

    def test_accessed_time(self):
        self.storage.save('file', ContentFile('test'))

        self.storage.open('file', 'r')
        accessed_time = self.storage.accessed_time('file')

        time.sleep(0.1)

        self.storage.open('file', 'r')
        new_accessed_time = self.storage.accessed_time('file')
        self.assertTrue(new_accessed_time > accessed_time)

    def test_created_time(self):
        self.storage.save('file', ContentFile('test'))
        created_time = self.storage.created_time('file')

        time.sleep(0.1)

        # make sure opening it doesn't change creation time.
        file = self.storage.open('file', 'r')

        after_open_created_time = self.storage.created_time('file')
        self.assertEqual(after_open_created_time, created_time)

        # make sure writing to it doesn't change creation time.
        file.write('test-test-test')
        self.storage.save('file', file)

        after_write_created_time = self.storage.created_time('file')
        self.assertEqual(after_write_created_time, created_time)

    @override_settings(INMEMORYSTORAGE_PERSIST=True)
    def test_persistance(self):
        storage_a = InMemoryStorage()
        storage_b = InMemoryStorage()

        # Can't use self.assertIs because it isn't available in Python 2.6
        self.assertTrue(storage_a.filesystem is storage_b.filesystem)

    def test_no_persistance_without_setting(self):
        storage_a = InMemoryStorage()
        storage_b = InMemoryStorage()

        # Can't use self.assertIs because it isn't available in Python 2.6
        self.assertTrue(storage_a.filesystem is not storage_b.filesystem)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

from setuptools import setup

requires = ['Django >= 1.4', 'six>=1.4.1']
tests_require = requires

setup(
    name="ambition-inmemorystorage",
    description="A non-persistent in-memory data storage backend for Django.",
    version="1.4.0",
    url="https://github.com/waveaccounting/dj-inmemorystorage",
    license=open('LICENSE').read(),
    long_description=open('README.rst').read(),
    author='Cody Soyland, Seán Hayes, Tore Birkeland, Nick Presta',
    author_email='opensource@waveapps.com',
    packages=[
        'inmemorystorage',
    ],
    zip_safe=True,
    install_requires=requires,
    tests_require=tests_require,
    test_suite='tests',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Internet :: WWW/HTTP',
    ]
)
